import sys
from dataclasses import dataclass
from unittest.mock import Mock

import pytest

from travertino.colors import hsl, hsla, rgb, rgba

if sys.version_info < (3, 10):
    _DATACLASS_KWARGS = {"init": False, "repr": False}
else:
    _DATACLASS_KWARGS = {"kw_only": True, "repr": False}


def apply_dataclass(cls):
    """Decorator to apply dataclass with arguments depending on Python version"""
    return dataclass(**_DATACLASS_KWARGS)(cls)


def mock_apply(cls):
    """Mock a style class's apply() method."""
    orig_init = cls.__init__

    def __init__(self, *args, **kwargs):
        self.apply = Mock()
        orig_init(self, *args, **kwargs)

    cls.__init__ = __init__
    return cls


def assert_equal_color(actual, expected, abs=1e-6):
    if {type(actual), type(expected)} == {rgba}:
        assert actual.rgba.r == pytest.approx(expected.rgba.r, abs=abs)
        assert actual.rgba.g == pytest.approx(expected.rgba.g, abs=abs)
        assert actual.rgba.b == pytest.approx(expected.rgba.b, abs=abs)
        assert actual.rgba.a == pytest.approx(expected.rgba.a, abs=abs)
    elif {type(actual), type(expected)} == {rgb}:
        assert actual.rgb.r == pytest.approx(expected.rgb.r, abs=abs)
        assert actual.rgb.g == pytest.approx(expected.rgb.g, abs=abs)
        assert actual.rgb.b == pytest.approx(expected.rgb.b, abs=abs)
    elif {type(actual), type(expected)} == {hsla}:
        assert actual.hsla.h == pytest.approx(expected.hsla.h, abs=abs)
        assert actual.hsla.s == pytest.approx(expected.hsla.s, abs=abs)
        assert actual.hsla.l == pytest.approx(expected.hsla.l, abs=abs)
        assert actual.hsla.a == pytest.approx(expected.hsla.a, abs=abs)
    elif {type(actual), type(expected)} == {hsl}:
        assert actual.hsl.h == pytest.approx(expected.hsl.h, abs=abs)
        assert actual.hsl.s == pytest.approx(expected.hsl.s, abs=abs)
        assert actual.hsl.l == pytest.approx(expected.hsl.l, abs=abs)
    else:
        raise ValueError(
            "Actual color and expected color should be of the same type. "
            f"But got actual:{type(actual)} and expected:{type(expected)}."
        )



from collections.abc import Sequence


class ImmutableList:
    def __init__(self, iterable):
        self._data = list(iterable)

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __eq__(self, other):
        return self._data == other

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)

    def __reversed__(self):
        return reversed(self._data)

    def index(self, value):
        return self._data.index(value)

    def count(self, value):
        return self._data.count(value)


Sequence.register(ImmutableList)



from __future__ import annotations

import re
from collections.abc import Sequence
from contextlib import nullcontext
from functools import partial
from unittest.mock import call
from warnings import catch_warnings, filterwarnings

import pytest

from travertino.properties.aliased import Condition, aliased_property
from travertino.properties.choices import Choices
from travertino.properties.immutablelist import ImmutableList
from travertino.properties.shorthand import directional_property
from travertino.properties.validated import list_property, validated_property
from travertino.style import BaseStyle

from .utils import apply_dataclass, mock_apply

VALUE1 = "value1"
VALUE2 = "value2"
VALUE3 = "value3"
VALUES = [VALUE1, VALUE2, VALUE3, None]


@mock_apply
@apply_dataclass
class Style(BaseStyle):
    # Some properties with explicit initial values
    explicit_const: str | int = validated_property(
        *VALUES, integer=True, initial=VALUE1
    )
    explicit_value: str | int = validated_property(*VALUES, integer=True, initial=0)
    explicit_none: str | int | None = validated_property(
        *VALUES, integer=True, initial=None
    )

    # A property with an implicit default value.
    # This usually means the default is platform specific.
    implicit: str | int | None = validated_property(
        VALUE1, VALUE2, VALUE3, integer=True
    )

    # A set of directional properties
    thing: tuple[str | int] | str | int = directional_property("thing{}")
    thing_top: str | int = validated_property(*VALUES, integer=True, initial=0)
    thing_right: str | int = validated_property(*VALUES, integer=True, initial=0)
    thing_bottom: str | int = validated_property(*VALUES, integer=True, initial=0)
    thing_left: str | int = validated_property(*VALUES, integer=True, initial=0)

    # Nothing below here needs to be tested in deprecated API:
    list_prop: list[str] = list_property(*VALUES, integer=True, initial=(VALUE2,))

    # A variety of aliases to other properties
    plain_alias: str | int = aliased_property(source="explicit_const")
    plain_alias_deprecated: str | int = aliased_property(
        source="explicit_const", deprecated=True
    )
    directional_alias: tuple[str | int] | str | int = aliased_property(source="thing")
    directional_alias_deprecated: tuple[str | int] | str | int = aliased_property(
        source="thing", deprecated=True
    )
    conditional_alias: str | int = aliased_property(
        source={
            Condition(thing_top=0): "explicit_const",
            Condition(thing_top=10, list_prop=[VALUE1, VALUE2]): "explicit_value",
            Condition(thing_top=10, list_prop=[VALUE1]): "explicit_none",
        }
    )
    conditional_alias_deprecated: str | int = aliased_property(
        source={
            Condition(thing_top=0): "explicit_const",
            Condition(thing_top=10, list_prop=[VALUE1, VALUE2]): "explicit_value",
            Condition(thing_top=10, list_prop=[VALUE1]): "explicit_none",
        },
        deprecated=True,
    )


VALUE_CHOICES = Choices(*VALUES, integer=True)

with catch_warnings():
    filterwarnings("ignore", category=DeprecationWarning)

    @mock_apply
    class DeprecatedStyle(BaseStyle):
        pass

    # Some properties with explicit initial values
    DeprecatedStyle.validated_property(
        "explicit_const", choices=VALUE_CHOICES, initial=VALUE1
    )
    DeprecatedStyle.validated_property(
        "explicit_value", choices=VALUE_CHOICES, initial=0
    )
    DeprecatedStyle.validated_property(
        "explicit_none", choices=VALUE_CHOICES, initial=None
    )

    # A property with an implicit default value.
    # This usually means the default is platform specific.
    DeprecatedStyle.validated_property(
        "implicit", choices=Choices(VALUE1, VALUE2, VALUE3, integer=True)
    )

    # A set of directional properties
    DeprecatedStyle.validated_property("thing_top", choices=VALUE_CHOICES, initial=0)
    DeprecatedStyle.validated_property("thing_right", choices=VALUE_CHOICES, initial=0)
    DeprecatedStyle.validated_property("thing_bottom", choices=VALUE_CHOICES, initial=0)
    DeprecatedStyle.validated_property("thing_left", choices=VALUE_CHOICES, initial=0)
    DeprecatedStyle.directional_property("thing%s")


class StyleSubclass(Style):
    pass


class DeprecatedStyleSubclass(DeprecatedStyle):
    pass


class Sibling(BaseStyle):
    pass


@mock_apply
@apply_dataclass
class MockedApplyStyle(BaseStyle):
    pass


def test_invalid_style():
    with pytest.raises(ValueError):
        # Define an invalid initial value on a validated property
        validated_property(*VALUES, integer=True, initial="something")

    with pytest.raises(ValueError):
        # Same for list property
        list_property(*VALUES, integer=True, initial=["something"])


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_positional_argument(StyleClass):
    # Could be the subclass or inherited __init__, depending on Python version / API
    # used.
    with pytest.raises(
        TypeError, match=r"__init__\(\) takes 1 positional argument but 2 were given"
    ):
        StyleClass(5)


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_constructor_invalid_property(StyleClass):
    """Whether dataclass or not, the error should be the same."""
    with pytest.raises(
        TypeError,
        match=r"Style\.__init__\(\) got an unexpected keyword argument 'bogus'",
    ):
        StyleClass(explicit_const=5, bogus=None)


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_create_and_copy(StyleClass):
    style = StyleClass(explicit_const=VALUE2, implicit=VALUE3)

    dup = style.copy()
    assert dup.explicit_const == VALUE2
    assert dup.explicit_value == 0
    assert dup.implicit == VALUE3


def test_deprecated_copy():
    style = MockedApplyStyle()

    with pytest.warns(DeprecationWarning):
        style_copy = style.copy(applicator=object())

    style_copy.apply.assert_called_once_with()


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_apply(StyleClass):
    style = StyleClass(explicit_const=VALUE2, implicit=VALUE3)
    assert style.apply.mock_calls == [call("explicit_const"), call("implicit")]
    style.apply.reset_mock()

    style.apply()
    style.apply.assert_called_once_with()


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_property_with_explicit_const(StyleClass):
    style = StyleClass()

    # Default value is VALUE1
    assert style.explicit_const is VALUE1
    style.apply.assert_not_called()

    # Modify the value
    style.explicit_const = 10

    assert style.explicit_const == 10
    style.apply.assert_called_once_with("explicit_const")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set the value to the same value.
    # No dirty notification is sent
    style.explicit_const = 10
    assert style.explicit_const == 10
    style.apply.assert_not_called()

    # Set the value to something new
    # A dirty notification is set.
    style.explicit_const = 20
    assert style.explicit_const == 20
    style.apply.assert_called_once_with("explicit_const")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear the property
    del style.explicit_const
    assert style.explicit_const is VALUE1
    style.apply.assert_called_once_with("explicit_const")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear the property again.
    # The underlying attribute won't exist, so this
    # should be a no-op.
    del style.explicit_const
    assert style.explicit_const is VALUE1
    style.apply.assert_not_called()


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_property_with_explicit_value(StyleClass):
    style = StyleClass()

    # Default value is 0
    assert style.explicit_value == 0
    style.apply.assert_not_called()

    # Modify the value
    style.explicit_value = 10

    assert style.explicit_value == 10
    style.apply.assert_called_once_with("explicit_value")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set the value to the same value.
    # No dirty notification is sent
    style.explicit_value = 10
    assert style.explicit_value == 10
    style.apply.assert_not_called()

    # Set the value to something new
    # A dirty notification is set.
    style.explicit_value = 20
    assert style.explicit_value == 20
    style.apply.assert_called_once_with("explicit_value")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear the property
    del style.explicit_value
    assert style.explicit_value == 0
    style.apply.assert_called_once_with("explicit_value")


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_property_with_explicit_none(StyleClass):
    style = StyleClass()

    # Default value is None
    assert style.explicit_none is None
    style.apply.assert_not_called()

    # Modify the value
    style.explicit_none = 10

    assert style.explicit_none == 10
    style.apply.assert_called_once_with("explicit_none")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set the property to the same value.
    # No dirty notification is sent
    style.explicit_none = 10
    assert style.explicit_none == 10
    style.apply.assert_not_called()

    # Set the property to something new
    # A dirty notification is set.
    style.explicit_none = 20
    assert style.explicit_none == 20
    style.apply.assert_called_once_with("explicit_none")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear the property
    del style.explicit_none
    assert style.explicit_none is None
    style.apply.assert_called_once_with("explicit_none")


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_property_with_implicit_default(StyleClass):
    style = StyleClass()

    # Default value is None
    assert style.implicit is None
    style.apply.assert_not_called()

    # Modify the value
    style.implicit = 10

    assert style.implicit == 10
    style.apply.assert_called_once_with("implicit")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set the value to the same value.
    # No dirty notification is sent
    style.implicit = 10
    assert style.implicit == 10
    style.apply.assert_not_called()

    # Set the value to something new
    # A dirty notification is set.
    style.implicit = 20
    assert style.implicit == 20
    style.apply.assert_called_once_with("implicit")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear the property
    del style.implicit
    assert style.implicit is None
    style.apply.assert_called_once_with("implicit")


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_set_initial_no_apply(StyleClass):
    """If a property hasn't been set, assigning it its initial value shouldn't apply."""
    style = StyleClass()

    # 0 is the initial value
    style.explicit_value = 0

    style.apply.assert_not_called()


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_directional_property(StyleClass):
    style = StyleClass()

    # Default value is 0
    assert style.thing == (0, 0, 0, 0)
    assert style.thing_top == 0
    assert style.thing_right == 0
    assert style.thing_bottom == 0
    assert style.thing_left == 0
    assert "thing" not in style
    style.apply.assert_not_called()

    # Set a value in one axis
    style.thing_top = 10

    assert style.thing == (10, 0, 0, 0)
    assert style.thing_top == 10
    assert style.thing_right == 0
    assert style.thing_bottom == 0
    assert style.thing_left == 0
    assert "thing" in style
    style.apply.assert_called_once_with("thing_top")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with a single item
    style.thing = (10,)

    assert style.thing == (10, 10, 10, 10)
    assert style.thing_top == 10
    assert style.thing_right == 10
    assert style.thing_bottom == 10
    assert style.thing_left == 10
    assert "thing" in style
    style.apply.assert_has_calls(
        [
            call("thing_right"),
            call("thing_bottom"),
            call("thing_left"),
        ]
    )

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with a single item
    style.thing = 30

    assert style.thing == (30, 30, 30, 30)
    assert style.thing_top == 30
    assert style.thing_right == 30
    assert style.thing_bottom == 30
    assert style.thing_left == 30
    assert "thing" in style
    style.apply.assert_has_calls(
        [
            call("thing_top"),
            call("thing_right"),
            call("thing_bottom"),
            call("thing_left"),
        ]
    )

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with 2 values
    style.thing = (10, 20)

    assert style.thing == (10, 20, 10, 20)
    assert style.thing_top == 10
    assert style.thing_right == 20
    assert style.thing_bottom == 10
    assert style.thing_left == 20
    assert "thing" in style
    style.apply.assert_has_calls(
        [
            call("thing_top"),
            call("thing_right"),
            call("thing_bottom"),
            call("thing_left"),
        ]
    )

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with 3 values
    style.thing = (10, 20, 30)

    assert style.thing == (10, 20, 30, 20)
    assert style.thing_top == 10
    assert style.thing_right == 20
    assert style.thing_bottom == 30
    assert style.thing_left == 20
    assert "thing" in style
    style.apply.assert_called_once_with("thing_bottom")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with 4 values
    style.thing = (10, 20, 30, 40)

    assert style.thing == (10, 20, 30, 40)
    assert style.thing_top == 10
    assert style.thing_right == 20
    assert style.thing_bottom == 30
    assert style.thing_left == 40
    assert "thing" in style
    style.apply.assert_called_once_with("thing_left")

    # Set a value directly with an invalid number of values
    with pytest.raises(ValueError):
        style.thing = ()

    with pytest.raises(ValueError):
        style.thing = (10, 20, 30, 40, 50)

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear a value on one axis
    del style.thing_top

    assert style.thing == (0, 20, 30, 40)
    assert style.thing_top == 0
    assert style.thing_right == 20
    assert style.thing_bottom == 30
    assert style.thing_left == 40
    assert "thing" in style
    style.apply.assert_called_once_with("thing_top")

    # Restore the top thing
    style.thing_top = 10

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear a value directly
    del style.thing

    assert style.thing == (0, 0, 0, 0)
    assert style.thing_top == 0
    assert style.thing_right == 0
    assert style.thing_bottom == 0
    assert style.thing_left == 0
    assert "thing" not in style
    style.apply.assert_has_calls(
        [
            call("thing_right"),
            call("thing_bottom"),
            call("thing_left"),
        ]
    )


@pytest.mark.parametrize(
    "value, expected",
    [
        ([VALUE1], [VALUE1]),
        (VALUE1, [VALUE1]),
        ([VALUE1, VALUE3], [VALUE1, VALUE3]),
        ([VALUE2, VALUE1], [VALUE2, VALUE1]),
        ([VALUE2, VALUE3, 1, 2, VALUE1], [VALUE2, VALUE3, 1, 2, VALUE1]),
        # Duplicates are kept, but "normalized" via validation.
        (
            [VALUE3, 1, VALUE3, "1", True, " 1", VALUE2],
            [VALUE3, 1, VALUE3, 1, 1, 1, VALUE2],
        ),
        # Other sequences should work too.
        ((VALUE1, VALUE3), [VALUE1, VALUE3]),
    ],
)
def test_list_property(value, expected):
    style = Style()
    style.list_prop = value
    assert style.list_prop == expected


@pytest.mark.parametrize(
    "value, error, match",
    [
        (
            5,
            TypeError,
            r"Value for list property list_prop must be a sequence\.",
        ),
        (
            # Fails because it's only a generator, not a comprehension:
            (i for i in [VALUE1, VALUE3]),
            TypeError,
            r"Value for list property list_prop must be a sequence.",
        ),
        (
            [VALUE3, VALUE1, "bogus"],
            ValueError,
            r"Invalid item value 'bogus' for list property list_prop; "
            r"Valid values are: none, value1, value2, value3, <integer>",
        ),
        (
            (),
            ValueError,
            r"List properties cannot be set to an empty sequence; "
            r"to reset a property, use del `style.list_prop`\.",
        ),
        (
            [],
            ValueError,
            r"List properties cannot be set to an empty sequence; "
            r"to reset a property, use del `style.list_prop`\.",
        ),
    ],
)
def test_list_property_invalid(value, error, match):
    style = Style()
    with pytest.raises(error, match=match):
        style.list_prop = value


def test_list_property_immutable():
    style = Style()
    style.list_prop = [1, 2, 3, VALUE2]
    prop = style.list_prop

    with pytest.raises(TypeError, match=r"does not support item assignment"):
        prop[0] = 5

    with pytest.raises(TypeError, match=r"doesn't support item deletion"):
        del prop[1]

    with pytest.raises(AttributeError):
        prop.insert(2, VALUE1)

    with pytest.raises(AttributeError):
        prop.append(VALUE3)

    with pytest.raises(AttributeError):
        prop.clear()

    with pytest.raises(AttributeError):
        prop.reverse()

    with pytest.raises(AttributeError):
        prop.pop()

    with pytest.raises(AttributeError):
        prop.remove(VALUE2)

    with pytest.raises(AttributeError):
        prop.extend([5, 6, 7])

    with pytest.raises(TypeError, match=r"unsupported operand type\(s\)"):
        prop += [4, 3, VALUE1]

    with pytest.raises(TypeError, match=r"unsupported operand type\(s\)"):
        prop += ImmutableList([4, 3, VALUE1])

    with pytest.raises(AttributeError):
        prop.sort()


def test_list_property_list_like():
    style = Style()
    style.list_prop = [1, 2, 3, VALUE2]
    prop = style.list_prop

    assert isinstance(prop, ImmutableList)
    assert prop == [1, 2, 3, VALUE2]
    assert prop == ImmutableList([1, 2, 3, VALUE2])
    assert prop[2] == 3
    assert str(prop) == repr(prop) == "[1, 2, 3, 'value2']"
    assert len(prop) == 4

    count = 0
    for _ in prop:
        count += 1
    assert count == 4

    assert [*reversed(prop)] == [VALUE2, 3, 2, 1]

    assert prop.index(3) == 2

    assert prop.count(VALUE2) == 1

    assert isinstance(prop, Sequence)


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_set_multiple_properties(StyleClass):
    style = StyleClass()

    # Set a pair of properties
    style.update(explicit_value=20, explicit_none=10)

    assert style.explicit_const is VALUE1
    assert style.explicit_none == 10
    assert style.explicit_value == 20
    style.apply.assert_has_calls(
        [
            call("explicit_value"),
            call("explicit_none"),
        ],
        any_order=True,
    )

    # Set a different pair of properties
    style.update(explicit_const=VALUE2, explicit_value=30)

    assert style.explicit_const is VALUE2
    assert style.explicit_value == 30
    assert style.explicit_none == 10
    style.apply.assert_has_calls(
        [
            call("explicit_const"),
            call("explicit_value"),
        ],
        any_order=True,
    )

    # Clear the applicator mock
    style.apply.reset_mock()

    # Setting a non-property
    with pytest.raises(NameError):
        style.update(not_a_property=10)

    style.apply.assert_not_called()


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_str(StyleClass):
    style = StyleClass()

    style.update(
        explicit_const=VALUE2,
        explicit_value=20,
        thing=(30, 40, 50, 60),
    )

    assert (
        str(style) == "explicit-const: value2; "
        "explicit-value: 20; "
        "thing-bottom: 50; "
        "thing-left: 60; "
        "thing-right: 40; "
        "thing-top: 30"
    )


def test_repr():
    # Doesn't need to be tested with deprecated API.
    style = Style(explicit_const=VALUE2, explicit_value=20, thing=(30, 40, 50, 60))

    assert repr(style) == (
        "Style("
        "explicit_const='value2', "
        "explicit_value=20, "
        "thing_bottom=50, "
        "thing_left=60, "
        "thing_right=40, "
        "thing_top=30"
        ")"
    )


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_dict(StyleClass):
    "Style declarations expose a dict-like interface"
    style = StyleClass()

    style.update(
        explicit_const=VALUE2,
        explicit_value=20,
        thing=(30, 40, 50, 60),
    )

    expected_keys = {
        "explicit_const",
        "explicit_value",
        "thing_bottom",
        "thing_left",
        "thing_right",
        "thing_top",
    }

    assert style.keys() == expected_keys

    assert sorted(style.items()) == sorted(
        [
            ("explicit_const", "value2"),
            ("explicit_value", 20),
            ("thing_bottom", 50),
            ("thing_left", 60),
            ("thing_right", 40),
            ("thing_top", 30),
        ]
    )

    # Style object has a length and is iterable.
    assert len(style) == 6
    for name in style:
        assert name in expected_keys

    # Properties that are set are in the keys.
    for name in expected_keys:
        assert name in style

    # Directional properties with one or more of the aliased properties set also count.
    assert "thing" in style

    # Valid properties that haven't been set are not in the keys.
    assert "implicit" not in style
    assert "explicit_none" not in style

    # Neither are invalid properties.
    assert "invalid_property" not in style

    # A property can be set, retrieved and cleared using the attribute name
    style["thing-bottom"] = 10
    assert style["thing-bottom"] == 10
    del style["thing-bottom"]
    assert style["thing-bottom"] == 0

    # A property can be set, retrieved and cleared using the Python attribute name
    style["thing_bottom"] = 10
    assert style["thing_bottom"] == 10
    del style["thing_bottom"]
    assert style["thing_bottom"] == 0

    # Property aliases can be accessed as well.
    style["thing"] = 5
    assert style["thing"] == (5, 5, 5, 5)
    del style["thing"]
    assert style["thing"] == (0, 0, 0, 0)

    # Clearing a valid property isn't an error
    del style["thing_bottom"]
    assert style["thing_bottom"] == 0

    # Non-existent properties raise KeyError
    with pytest.raises(KeyError):
        style["no-such-property"] = "no-such-value"

    with pytest.raises(KeyError):
        style["no-such-property"]

    with pytest.raises(KeyError):
        del style["no-such-property"]


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_set_to_initial(StyleClass):
    """A property set to its initial value is distinct from an unset property."""
    style = StyleClass()

    # explicit_const's initial value is VALUE1.
    assert style.explicit_const == VALUE1
    assert "explicit_const" not in style

    # The unset property shouldn't affect the value when overlaid over a style with
    # that property set.
    non_initial_style = StyleClass(explicit_const=VALUE2)
    union = non_initial_style | style
    assert union.explicit_const == VALUE2
    assert "explicit_const" in union

    # The property should count as set, even when set to the same initial value.
    style.explicit_const = VALUE1
    assert style.explicit_const == VALUE1
    assert "explicit_const" in style

    # The property should now overwrite.
    union = non_initial_style | style
    assert union.explicit_const == VALUE1
    assert "explicit_const" in union


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
@pytest.mark.parametrize("instantiate", [True, False])
def test_union_operators(StyleClass, instantiate):
    """Styles support | and |= with dicts and with their own class."""
    left = StyleClass(explicit_value=VALUE1, implicit=VALUE2)

    style_dict = {"thing_top": 5, "implicit": VALUE3}
    right = StyleClass(**style_dict) if instantiate else style_dict

    # Standard operator
    result = left | right

    # Original objects unchanged
    assert left["explicit_value"] == VALUE1
    assert left["implicit"] == VALUE2

    assert right["thing_top"] == 5
    assert right["implicit"] == VALUE3

    # Unshared properties assigned
    assert result["explicit_const"] == VALUE1
    assert result["thing_top"] == 5

    # Common property overridden by second operand
    assert result["implicit"] == VALUE3

    # In-place version
    left |= right

    # Common property updated on lefthand
    assert left["explicit_value"] == VALUE1
    assert left["implicit"] == VALUE3

    # Righthand unchanged
    assert right["thing_top"] == 5
    assert right["implicit"] == VALUE3


@pytest.mark.parametrize(
    "StyleClass, OtherClass",
    [
        (Style, StyleSubclass),
        (Style, Sibling),
        (Style, int),
        (Style, list),
        (DeprecatedStyle, DeprecatedStyleSubclass),
        (DeprecatedStyle, Sibling),
        (DeprecatedStyle, int),
        (DeprecatedStyle, list),
    ],
)
def test_union_operators_invalid_type(StyleClass, OtherClass):
    """Styles do not support | or |= with other style classes or with non-mappings."""

    left = StyleClass()
    right = OtherClass()

    with pytest.raises(TypeError, match=r"unsupported operand type"):
        left | right

    with pytest.raises(TypeError, match=r"unsupported operand type"):
        left |= right


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
@pytest.mark.parametrize(
    "right, error",
    [
        ({"implicit": "bogus_value"}, ValueError),
        ({"bogus_key": 3.12}, NameError),
    ],
)
def test_union_operators_invalid_key_value(StyleClass, right, error):
    """Operators will accept any mapping, but invalid keys/values are still an error."""
    left = StyleClass()

    with pytest.raises(error):
        left | right

    with pytest.raises(error):
        left |= right


def test_deprecated_class_methods():
    """Toga < 0.5.0 can still define style properties with the old API."""

    class OldStyle(BaseStyle):
        pass

    with pytest.warns(DeprecationWarning):
        OldStyle.validated_property(
            "implicit", Choices(*VALUES, integer=True), initial=VALUE1
        )

    with pytest.warns(DeprecationWarning):
        OldStyle.directional_property("thing%s")


def test_deprecated_reapply():
    """Reapply() is deprecated (but still calls apply()."""
    style = Style()
    with pytest.warns(DeprecationWarning):
        style.reapply()

    style.apply.assert_called_once_with()


def test_deprecated_import():
    """Toga < 0.5.0 can still import what it needs from Travertino."""
    with pytest.deprecated_call():
        from travertino.declaration import BaseStyle, Choices  # noqa


@pytest.mark.parametrize(
    "source, alias, initial, value, context",
    [
        (
            "explicit_const",
            "plain_alias",
            VALUE1,
            VALUE3,
            nullcontext,
        ),
        (
            "explicit_const",
            "plain_alias_deprecated",
            VALUE1,
            VALUE3,
            partial(
                pytest.deprecated_call,
                match=(
                    r"Style\.plain_alias_deprecated is deprecated\. "
                    r"Use Style\.explicit_const instead\."
                ),
            ),
        ),
        # There *shouldn't* be anything special about an alias to a
        # directional_property vs. to a validated_property, but it doesn't hurt to
        # check.
        (
            "thing",
            "directional_alias",
            (0, 0, 0, 0),
            (10, 5, 10, 5),
            nullcontext,
        ),
        (
            "thing",
            "directional_alias_deprecated",
            (0, 0, 0, 0),
            (10, 5, 10, 5),
            partial(
                pytest.deprecated_call,
                match=(
                    r"Style\.directional_alias_deprecated is deprecated\. "
                    r"Use Style\.thing instead\."
                ),
            ),
        ),
    ],
)
def test_simple_alias(source, alias, initial, value, context):
    """An alias with a constant source refers consistently to the source property."""
    style = Style()

    # Set on original, check on alias
    with context():
        assert alias not in style

    style[source] = value

    with context():
        assert alias in style
    with context():
        assert style[alias] == value

    del style[source]

    with context():
        assert alias not in style
    with context():
        assert style[alias] == initial

    # Set on alias, check on original.
    assert source not in style

    with context():
        style[alias] = value

    assert source in style
    assert style[source] == value

    with context():
        del style[alias]

    assert source not in style
    assert style[source] == initial


@pytest.mark.parametrize(
    "alias, context",
    [
        ("conditional_alias", nullcontext),
        ("conditional_alias_deprecated", pytest.deprecated_call),
    ],
)
@pytest.mark.parametrize(
    "properties, source, value",
    [
        ({}, "explicit_const", VALUE1),
        ({"thing_top": 10, "list_prop": [VALUE1, VALUE2]}, "explicit_value", 0),
        ({"thing_top": 10, "list_prop": [VALUE1]}, "explicit_none", None),
    ],
)
def test_conditional_alias(alias, context, properties, source, value):
    """An alias correctly checks conditions to determine which property to access."""
    style = Style(**properties)
    with context():
        assert style[alias] == style[source] == value


@pytest.mark.parametrize(
    "alias",
    # No deprecation warnings are given; there's no way to even tell which newer name
    # should be instead.
    ["conditional_alias", "conditional_alias_deprecated"],
)
@pytest.mark.parametrize(
    "properties",
    [
        {"thing_top": 1},
        {"thing_top": 10, "list_prop": [VALUE1, VALUE3]},
        {"thing_top": 11},
    ],
)
def test_conditional_alias_invald(alias, properties):
    """If no condition is valid, using the alias raises an AttributeError."""
    error_msg = re.escape(
        f"'{alias}' is only supported when (thing_top == 0) or (thing_top == 10; "
        "list_prop == ['value1', 'value2']) or (thing_top == 10; list_prop == "
        "['value1'])"
    )
    style = Style(**properties)
    with pytest.raises(AttributeError, match=error_msg):
        style[alias]

    with pytest.raises(AttributeError, match=error_msg):
        style[alias] = 15

    with pytest.raises(AttributeError, match=error_msg):
        alias in style

    with pytest.raises(AttributeError, match=error_msg):
        del style[alias]


@pytest.mark.parametrize(
    "alias, context",
    [
        ("conditional_alias", nullcontext),
        ("conditional_alias_deprecated", pytest.deprecated_call),
    ],
)
@pytest.mark.parametrize(
    "properties",
    [
        {"alias": 1, "thing_top": 10, "list_prop": [VALUE1]},
        {"thing_top": 10, "alias": 1, "list_prop": [VALUE1]},
        {"thing_top": 10, "list_prop": [VALUE1], "alias": 1},
    ],
)
def test_conditional_alias_simultaneous_setting(alias, context, properties):
    """Setting an alias along with its conditions works, in any order."""
    properties = {
        (alias if name == "alias" else name): value
        for name, value in properties.items()
    }
    with context():
        style = Style(**properties)
    with context():
        assert style[alias] == 1

    style = Style()
    with context():
        style.update(**properties)
    with context():
        assert style[alias] == 1



import pytest

from travertino.colors import color, hsl, hsla, rgb, rgba


def assert_equal_hsl(value, expected):
    # Nothing fancy - a color is equal if the attributes are all the same
    actual = color(value)
    assert actual.h == expected.h
    assert actual.s == expected.s
    assert actual.l == expected.l
    assert actual.a == pytest.approx(expected.a, abs=0.001)


def assert_equal_rgb(value, expected):
    # Nothing fancy - a color is equal if the attributes are all the same
    actual = color(value)
    assert actual.r == expected.r
    assert actual.g == expected.g
    assert actual.b == expected.b
    assert actual.a == pytest.approx(expected.a, abs=0.001)


def test_noop():
    assert_equal_rgb(rgba(1, 2, 3, 0.5), rgba(1, 2, 3, 0.5))
    assert_equal_hsl(hsl(1, 0.2, 0.3), hsl(1, 0.2, 0.3))


@pytest.mark.parametrize(
    "value, expected",
    [
        ("rgb(1,2,3)", (1, 2, 3)),
        ("rgb(1, 2, 3)", (1, 2, 3)),
        ("rgb( 1 , 2 , 3)", (1, 2, 3)),
        ("#123", (0x11, 0x22, 0x33)),
        ("#112233", (0x11, 0x22, 0x33)),
        ("#abc", (0xAA, 0xBB, 0xCC)),
        ("#ABC", (0xAA, 0xBB, 0xCC)),
        ("#abcdef", (0xAB, 0xCD, 0xEF)),
        ("#ABCDEF", (0xAB, 0xCD, 0xEF)),
    ],
)
def test_rgb(value, expected):
    assert_equal_rgb(value, rgb(*expected))


@pytest.mark.parametrize(
    "value",
    [
        "10, 20",
        "a, 10, 20",
        "10, b, 20",
        "10, 20, c",
        "10, 20, 30, 0.5",
    ],
)
def test_rgb_invalid(value):
    with pytest.raises(ValueError):
        color(f"rgb({value})")


@pytest.mark.parametrize(
    "value, expected",
    [
        ("rgba(1,2,3,0.5)", (1, 2, 3, 0.5)),
        ("rgba(1, 2, 3, 0.5)", (1, 2, 3, 0.5)),
        ("rgba( 1 , 2 , 3 , 0.5)", (1, 2, 3, 0.5)),
        ("#1234", (0x11, 0x22, 0x33, 0.2666)),
        ("#11223344", (0x11, 0x22, 0x33, 0.2666)),
        ("#abcd", (0xAA, 0xBB, 0xCC, 0.8666)),
        ("#ABCD", (0xAA, 0xBB, 0xCC, 0.8666)),
        ("#abcdefba", (0xAB, 0xCD, 0xEF, 0.7294)),
        ("#ABCDEFBA", (0xAB, 0xCD, 0xEF, 0.7294)),
    ],
)
def test_rgba(value, expected):
    assert_equal_rgb(value, rgba(*expected))


@pytest.mark.parametrize(
    "value",
    [
        "10, 20, 30",
        "a, 10, 20, 0.5",
        "10, b, 20, 0.5",
        "10, 20, c, 0.5",
        "10, 20, 30, c",
        "10, 20, 30, 0.5, 5",
    ],
)
def test_rgba_invalid(value):
    with pytest.raises(ValueError):
        color(f"rgba({value})")


@pytest.mark.parametrize(
    "value",
    [
        "1,20%,30%",
        "1, 20%, 30%",
        "1, 20% , 30%",
    ],
)
def test_hsl(value):
    assert_equal_hsl(f"hsl({value})", hsl(1, 0.2, 0.3))


@pytest.mark.parametrize(
    "value",
    [
        "1, 20%",
        "a, 20%, 30%",
        "1, a, 30%",
        "1, 20%, a)",
        "1, 20%, 30%, 0.5)",
    ],
)
def test_hsl_invalid(value):
    with pytest.raises(ValueError):
        color(f"hsl({value}")


@pytest.mark.parametrize(
    "value",
    [
        "1,20%,30%,0.5",
        "1, 20%, 30%, 0.5",
        " 1, 20% , 30% , 0.5",
    ],
)
def test_hsla(value):
    assert_equal_hsl(f"hsla({value})", hsla(1, 0.2, 0.3, 0.5))


@pytest.mark.parametrize(
    "value",
    [
        "1, 20%, 30%",
        "a, 20%, 30%, 0.5",
        "1, a, 30%, 0.5",
        "1, 20%, a, 0.5",
        "1, 20%, 30%, a",
        "1, 20%, 30%, 0.5, 5",
    ],
)
def test_hsla_invalid(value):
    with pytest.raises(ValueError):
        color(f"hsla({value})")


@pytest.mark.parametrize(
    "value, expected",
    [
        ("Red", (0xFF, 0, 0)),
        ("RED", (0xFF, 0, 0)),
        ("red", (0xFF, 0, 0)),
        ("rEd", (0xFF, 0, 0)),
        ("CornflowerBlue", (0x64, 0x95, 0xED)),
        ("cornflowerblue", (0x64, 0x95, 0xED)),
        ("CORNFLOWERBLUE", (0x64, 0x95, 0xED)),
        ("Cornflowerblue", (0x64, 0x95, 0xED)),
        ("CoRnFlOwErBlUe", (0x64, 0x95, 0xED)),
    ],
)
def test_named_color(value, expected):
    assert_equal_rgb(value, rgb(*expected))


def test_named_color_invalid():
    with pytest.raises(ValueError):
        color("not a color")


@pytest.mark.parametrize(
    "num_digits",
    [1, 2, 5, 7, 9, 10],
)
def test_hash_mark_invalid(num_digits):
    """An invalid number of digts after # raises a ValueError."""
    with pytest.raises(ValueError):
        color(f"#{'1' * num_digits}")



from __future__ import annotations

# flake8: NOQA: F405
from .constants import *


class Color:
    "A base class for all colorspace representations."

    def __eq__(self, other):
        try:
            c1 = self.rgba
            c2 = other.rgba

            return c1.r == c2.r and c1.g == c2.g and c1.b == c2.b and c1.a == c2.a
        except AttributeError:
            return False

    @classmethod
    def _validate_between(cls, content_name, value, min_value, max_value):
        if value < min_value or value > max_value:
            raise ValueError(
                "{} value should be between {}-{}. Got {}".format(
                    content_name, min_value, max_value, value
                )
            )

    @classmethod
    def _validate_partial(cls, content_name, value):
        cls._validate_between(content_name, value, 0, 1)

    @classmethod
    def _validate_alpha(cls, value):
        cls._validate_partial("alpha", value)

    def blend_over(
        self,
        back_color: Color,
    ) -> rgba:
        """Performs the "over" straight alpha blending operation, compositing
        the front color over the back color.

        **Straight alpha blending** is not the same as
        **Premultiplied alpha blending**, see:
        https://en.wikipedia.org/wiki/Alpha_compositing#Straight_versus_premultiplied

        :param back_color: The background color.

        :returns: The blended color.
        """
        # The blending operation implemented here is the "over" operation and
        # replicates CSS's rgba mechanism. For the formulae used here, see:
        # https://en.wikipedia.org/wiki/Alpha_compositing#Description

        # Convert the input colors to rgba in order to do the calculation.
        front_color = self.rgba
        back_color = back_color.rgba

        if front_color.a == 1:
            # If the front color is fully opaque then the result will be the same as
            # front color.
            return front_color.rgba

        blended_alpha = min(
            1, max(0, (front_color.a + ((1 - front_color.a) * back_color.a)))
        )

        if blended_alpha == 0:
            # Don't further blend the color, to prevent divide by 0.
            return rgba(0, 0, 0, 0)
        else:
            blended_color = rgba(
                # Red Component
                min(
                    255,
                    max(
                        0,
                        round(
                            (
                                (front_color.r * front_color.a)
                                + (back_color.r * back_color.a * (1 - front_color.a))
                            )
                            / blended_alpha
                        ),
                    ),
                ),
                # Green Component
                min(
                    255,
                    max(
                        0,
                        round(
                            (
                                (front_color.g * front_color.a)
                                + (back_color.g * back_color.a * (1 - front_color.a))
                            )
                            / blended_alpha
                        ),
                    ),
                ),
                # Blue Component
                min(
                    255,
                    max(
                        0,
                        round(
                            (
                                (front_color.b * front_color.a)
                                + (back_color.b * back_color.a * (1 - front_color.a))
                            )
                            / blended_alpha
                        ),
                    ),
                ),
                # Alpha component
                min(1, max(0, blended_alpha)),
            )
            return blended_color

    def unblend_over(self, back_color: Color, front_color_alpha: float) -> rgba:
        """Performs the reverse of the "over" straight alpha blending operation,
        returning the front color.

        Note: Unblending of blended colors might produce front color with slightly
        imprecise component values compared to the original front color. This is
        due to the loss of some amount of precision during the calculation and
        conversion process between the different color formats. For example,
        unblending of a hsla blended color might might produce a slightly imprecise
        original front color, since the alpha blending and unblending is calculated
        after conversion to rgba values.

        :param back_color: The background color.
        :param front_color_alpha: The original alpha value of the front color,
            within the range of (0, 1].

        :raises ValueError: If the value of :any:`front_color_alpha` is not within
            the range of (0, 1]. The value cannot be 0, since the blended color produced
            will be equal to the back color, and all information related to the front
            color will be lost.

        :returns: The original front color.
        """
        # The blending operation implemented here is the reverse of the "over"
        # operation and replicates CSS's rgba mechanism. The formula used here
        # are derived from the "over" straight alpha blending operation formula,
        # see: https://en.wikipedia.org/wiki/Alpha_compositing#Description

        blended_color = self.rgba
        back_color = back_color.rgba
        if not (0 < front_color_alpha <= 1):
            raise ValueError(
                "The value of front_color_alpha must be within the range of (0, 1]."
            )
        else:
            front_color = rgba(
                # Red Component
                min(
                    255,
                    max(
                        0,
                        round(
                            (
                                (blended_color.r * blended_color.a)
                                - (
                                    back_color.r
                                    * back_color.a
                                    * (1 - front_color_alpha)
                                )
                            )
                            / front_color_alpha
                        ),
                    ),
                ),
                # Green Component
                min(
                    255,
                    max(
                        0,
                        round(
                            (
                                (blended_color.g * blended_color.a)
                                - (
                                    back_color.g
                                    * back_color.a
                                    * (1 - front_color_alpha)
                                )
                            )
                            / front_color_alpha
                        ),
                    ),
                ),
                # Blue Component
                min(
                    255,
                    max(
                        0,
                        round(
                            (
                                (blended_color.b * blended_color.a)
                                - (
                                    back_color.b
                                    * back_color.a
                                    * (1 - front_color_alpha)
                                )
                            )
                            / front_color_alpha
                        ),
                    ),
                ),
                # Alpha Component
                front_color_alpha,
            )
            return front_color.rgba


class rgba(Color):
    "A representation of an RGBA color."

    def __init__(self, r, g, b, a):
        self._validate_rgb("red", r)
        self._validate_rgb("green", g)
        self._validate_rgb("blue", b)
        self._validate_alpha(a)
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    def __hash__(self):
        return hash(("RGBA-color", self.r, self.g, self.b, self.a))

    def __repr__(self):
        return f"rgba({self.r}, {self.g}, {self.b}, {self.a})"

    @classmethod
    def _validate_rgb(cls, content_name, value):
        cls._validate_between(content_name, value, 0, 255)

    @property
    def rgba(self):
        # Explicitly return a rgba value, to ensure that classes which inherit the
        # parent rgba class, actually return the rgba() value, instead of the child
        # class. For example, without this fix, doing rgb(20, 20, 20).rgba will
        # return rgb(20, 20, 20), instead of rgba(20, 20, 20, 1).
        return rgba(self.r, self.g, self.b, self.a)

    @property
    def rgb(self):
        return rgb(self.r, self.g, self.b)

    @property
    def hsla(self) -> hsla:
        # Formula used here is from: https://en.wikipedia.org/wiki/HSL_and_HSV#From_RGB
        r_prime = self.r / 255
        g_prime = self.g / 255
        b_prime = self.b / 255

        max_component = max(r_prime, g_prime, b_prime)
        min_component = min(r_prime, g_prime, b_prime)
        value = max_component
        chroma = max_component - min_component

        lightness = (max_component + min_component) / 2

        if chroma == 0:
            hue = 0
        elif value == r_prime:
            hue = 60 * (((g_prime - b_prime) / chroma) % 6)
        elif value == g_prime:
            hue = 60 * (((b_prime - r_prime) / chroma) + 2)
        else:  # value == b_prime:
            hue = 60 * (((r_prime - g_prime) / chroma) + 4)

        if lightness in {0, 1}:
            saturation = 0
        else:
            saturation = chroma / (1 - abs((2 * value) - chroma - 1))

        return hsla(
            hue % 360,  # [0,360)
            min(1, max(0, saturation)),  # [0,1]
            min(1, max(0, lightness)),  # [0,1]
            min(1, max(0, self.a)),  # [0,1]
        )

    @property
    def hsl(self):
        return self.hsla.hsl


class rgb(rgba):
    "A representation of an RGB color"

    def __init__(self, r, g, b):
        super().__init__(r, g, b, 1.0)

    def __repr__(self):
        return f"rgb({self.r}, {self.g}, {self.b})"


class hsla(Color):
    "A representation of an HSLA color."

    def __init__(self, h, s, l, a=1.0):
        self._validate_between("hue", h, 0, 360)
        self._validate_partial("saturation", s)
        self._validate_partial("lightness", l)
        self._validate_alpha(a)
        self.h = h
        self.s = s
        self.l = l  # NOQA; E741
        self.a = a

    def __hash__(self):
        return hash(("HSLA-color", self.h, self.s, self.l, self.a))

    def __repr__(self):
        return f"hsla({self.h}, {self.s}, {self.l}, {self.a})"

    @property
    def hsla(self):
        # Explicitly return a hsla value, to ensure that classes which inherit the
        # parent hsla class, actually return the hsla() value, instead of the child
        # class. For example, without this fix, doing hsl(210, 1, 0.5).hsla will
        # return hsl(210, 1, 0.5), instead of hsla(210, 1, 0.5, 1).
        return hsla(self.h, self.s, self.l, self.a)

    @property
    def hsl(self):
        return hsl(self.h, self.s, self.l)

    @property
    def rgba(self):
        c = (1.0 - abs(2.0 * self.l - 1.0)) * self.s
        h = self.h / 60.0
        x = c * (1.0 - abs(h % 2 - 1.0))
        m = self.l - 0.5 * c

        if h < 1.0:
            r, g, b = c + m, x + m, m
        elif h < 2.0:
            r, g, b = x + m, c + m, m
        elif h < 3.0:
            r, g, b = m, c + m, x + m
        elif h < 4.0:
            r, g, b = m, x + m, c + m
        elif h < 5.0:
            r, g, b = m, x + m, c + m
        else:
            r, g, b = c + m, m, x + m

        return rgba(
            round(r * 0xFF),
            round(g * 0xFF),
            round(b * 0xFF),
            self.a,
        )

    @property
    def rgb(self):
        return self.rgba.rgb


class hsl(hsla):
    "A representation of an HSL color"

    def __init__(self, h, s, l):
        super().__init__(h, s, l, 1.0)

    def __repr__(self):
        return f"hsl({self.h}, {self.s}, {self.l})"


def color(value):
    """Parse a color from a value.

    Accepts:
    * rgb() instances
    * hsl() instances
    * '#rgb'
    * '#rgba'
    * '#rrggbb'
    * '#rrggbbaa'
    * '#RGB'
    * '#RGBA'
    * '#RRGGBB'
    * '#RRGGBBAA'
    * 'rgb(0, 0, 0)'
    * 'rgba(0, 0, 0, 0.0)'
    * 'hsl(0, 0%, 0%)'
    * 'hsla(0, 0%, 0%, 0.0)'
    * A named color
    """

    if isinstance(value, Color):
        return value

    elif isinstance(value, str):
        if value[0] == "#":
            if len(value) == 4:
                return rgb(
                    r=int(value[1] + value[1], 16),
                    g=int(value[2] + value[2], 16),
                    b=int(value[3] + value[3], 16),
                )
            elif len(value) == 5:
                return rgba(
                    r=int(value[1] + value[1], 16),
                    g=int(value[2] + value[2], 16),
                    b=int(value[3] + value[3], 16),
                    a=int(value[4] + value[4], 16) / 0xFF,
                )
            elif len(value) == 7:
                return rgb(
                    r=int(value[1:3], 16),
                    g=int(value[3:5], 16),
                    b=int(value[5:7], 16),
                )
            elif len(value) == 9:
                return rgba(
                    r=int(value[1:3], 16),
                    g=int(value[3:5], 16),
                    b=int(value[5:7], 16),
                    a=int(value[7:9], 16) / 0xFF,
                )
        elif value.startswith("rgba"):
            try:
                values = value[5:-1].split(",")
                if len(values) == 4:
                    return rgba(
                        int(values[0]),
                        int(values[1]),
                        int(values[2]),
                        float(
                            values[3],
                        ),
                    )
            except ValueError:
                pass
        elif value.startswith("rgb"):
            try:
                values = value[4:-1].split(",")
                if len(values) == 3:
                    return rgb(
                        int(values[0]),
                        int(values[1]),
                        int(values[2]),
                    )
            except ValueError:
                pass

        elif value.startswith("hsla"):
            try:
                values = value[5:-1].split(",")
                if len(values) == 4:
                    return hsla(
                        int(values[0]),
                        int(values[1].strip().rstrip("%")) / 100.0,
                        int(values[2].strip().rstrip("%")) / 100.0,
                        float(values[3]),
                    )
            except ValueError:
                pass

        elif value.startswith("hsl"):
            try:
                values = value[4:-1].split(",")
                if len(values) == 3:
                    return hsl(
                        int(values[0]),
                        int(values[1].strip().rstrip("%")) / 100.0,
                        int(values[2].strip().rstrip("%")) / 100.0,
                    )
            except ValueError:
                pass
        else:
            try:
                return NAMED_COLOR[value.lower()]
            except KeyError:
                pass

    raise ValueError("Unknown color %s" % value)


NAMED_COLOR = {
    ALICEBLUE: rgb(0xF0, 0xF8, 0xFF),
    ANTIQUEWHITE: rgb(0xFA, 0xEB, 0xD7),
    AQUA: rgb(0x00, 0xFF, 0xFF),
    AQUAMARINE: rgb(0x7F, 0xFF, 0xD4),
    AZURE: rgb(0xF0, 0xFF, 0xFF),
    BEIGE: rgb(0xF5, 0xF5, 0xDC),
    BISQUE: rgb(0xFF, 0xE4, 0xC4),
    BLACK: rgb(0x00, 0x00, 0x00),
    BLANCHEDALMOND: rgb(0xFF, 0xEB, 0xCD),
    BLUE: rgb(0x00, 0x00, 0xFF),
    BLUEVIOLET: rgb(0x8A, 0x2B, 0xE2),
    BROWN: rgb(0xA5, 0x2A, 0x2A),
    BURLYWOOD: rgb(0xDE, 0xB8, 0x87),
    CADETBLUE: rgb(0x5F, 0x9E, 0xA0),
    CHARTREUSE: rgb(0x7F, 0xFF, 0x00),
    CHOCOLATE: rgb(0xD2, 0x69, 0x1E),
    CORAL: rgb(0xFF, 0x7F, 0x50),
    CORNFLOWERBLUE: rgb(0x64, 0x95, 0xED),
    CORNSILK: rgb(0xFF, 0xF8, 0xDC),
    CRIMSON: rgb(0xDC, 0x14, 0x3C),
    CYAN: rgb(0x00, 0xFF, 0xFF),
    DARKBLUE: rgb(0x00, 0x00, 0x8B),
    DARKCYAN: rgb(0x00, 0x8B, 0x8B),
    DARKGOLDENROD: rgb(0xB8, 0x86, 0x0B),
    DARKGRAY: rgb(0xA9, 0xA9, 0xA9),
    DARKGREY: rgb(0xA9, 0xA9, 0xA9),
    DARKGREEN: rgb(0x00, 0x64, 0x00),
    DARKKHAKI: rgb(0xBD, 0xB7, 0x6B),
    DARKMAGENTA: rgb(0x8B, 0x00, 0x8B),
    DARKOLIVEGREEN: rgb(0x55, 0x6B, 0x2F),
    DARKORANGE: rgb(0xFF, 0x8C, 0x00),
    DARKORCHID: rgb(0x99, 0x32, 0xCC),
    DARKRED: rgb(0x8B, 0x00, 0x00),
    DARKSALMON: rgb(0xE9, 0x96, 0x7A),
    DARKSEAGREEN: rgb(0x8F, 0xBC, 0x8F),
    DARKSLATEBLUE: rgb(0x48, 0x3D, 0x8B),
    DARKSLATEGRAY: rgb(0x2F, 0x4F, 0x4F),
    DARKSLATEGREY: rgb(0x2F, 0x4F, 0x4F),
    DARKTURQUOISE: rgb(0x00, 0xCE, 0xD1),
    DARKVIOLET: rgb(0x94, 0x00, 0xD3),
    DEEPPINK: rgb(0xFF, 0x14, 0x93),
    DEEPSKYBLUE: rgb(0x00, 0xBF, 0xFF),
    DIMGRAY: rgb(0x69, 0x69, 0x69),
    DIMGREY: rgb(0x69, 0x69, 0x69),
    DODGERBLUE: rgb(0x1E, 0x90, 0xFF),
    FIREBRICK: rgb(0xB2, 0x22, 0x22),
    FLORALWHITE: rgb(0xFF, 0xFA, 0xF0),
    FORESTGREEN: rgb(0x22, 0x8B, 0x22),
    FUCHSIA: rgb(0xFF, 0x00, 0xFF),
    GAINSBORO: rgb(0xDC, 0xDC, 0xDC),
    GHOSTWHITE: rgb(0xF8, 0xF8, 0xFF),
    GOLD: rgb(0xFF, 0xD7, 0x00),
    GOLDENROD: rgb(0xDA, 0xA5, 0x20),
    GRAY: rgb(0x80, 0x80, 0x80),
    GREY: rgb(0x80, 0x80, 0x80),
    GREEN: rgb(0x00, 0x80, 0x00),
    GREENYELLOW: rgb(0xAD, 0xFF, 0x2F),
    HONEYDEW: rgb(0xF0, 0xFF, 0xF0),
    HOTPINK: rgb(0xFF, 0x69, 0xB4),
    INDIANRED: rgb(0xCD, 0x5C, 0x5C),
    INDIGO: rgb(0x4B, 0x00, 0x82),
    IVORY: rgb(0xFF, 0xFF, 0xF0),
    KHAKI: rgb(0xF0, 0xE6, 0x8C),
    LAVENDER: rgb(0xE6, 0xE6, 0xFA),
    LAVENDERBLUSH: rgb(0xFF, 0xF0, 0xF5),
    LAWNGREEN: rgb(0x7C, 0xFC, 0x00),
    LEMONCHIFFON: rgb(0xFF, 0xFA, 0xCD),
    LIGHTBLUE: rgb(0xAD, 0xD8, 0xE6),
    LIGHTCORAL: rgb(0xF0, 0x80, 0x80),
    LIGHTCYAN: rgb(0xE0, 0xFF, 0xFF),
    LIGHTGOLDENRODYELLOW: rgb(0xFA, 0xFA, 0xD2),
    LIGHTGRAY: rgb(0xD3, 0xD3, 0xD3),
    LIGHTGREY: rgb(0xD3, 0xD3, 0xD3),
    LIGHTGREEN: rgb(0x90, 0xEE, 0x90),
    LIGHTPINK: rgb(0xFF, 0xB6, 0xC1),
    LIGHTSALMON: rgb(0xFF, 0xA0, 0x7A),
    LIGHTSEAGREEN: rgb(0x20, 0xB2, 0xAA),
    LIGHTSKYBLUE: rgb(0x87, 0xCE, 0xFA),
    LIGHTSLATEGRAY: rgb(0x77, 0x88, 0x99),
    LIGHTSLATEGREY: rgb(0x77, 0x88, 0x99),
    LIGHTSTEELBLUE: rgb(0xB0, 0xC4, 0xDE),
    LIGHTYELLOW: rgb(0xFF, 0xFF, 0xE0),
    LIME: rgb(0x00, 0xFF, 0x00),
    LIMEGREEN: rgb(0x32, 0xCD, 0x32),
    LINEN: rgb(0xFA, 0xF0, 0xE6),
    MAGENTA: rgb(0xFF, 0x00, 0xFF),
    MAROON: rgb(0x80, 0x00, 0x00),
    MEDIUMAQUAMARINE: rgb(0x66, 0xCD, 0xAA),
    MEDIUMBLUE: rgb(0x00, 0x00, 0xCD),
    MEDIUMORCHID: rgb(0xBA, 0x55, 0xD3),
    MEDIUMPURPLE: rgb(0x93, 0x70, 0xDB),
    MEDIUMSEAGREEN: rgb(0x3C, 0xB3, 0x71),
    MEDIUMSLATEBLUE: rgb(0x7B, 0x68, 0xEE),
    MEDIUMSPRINGGREEN: rgb(0x00, 0xFA, 0x9A),
    MEDIUMTURQUOISE: rgb(0x48, 0xD1, 0xCC),
    MEDIUMVIOLETRED: rgb(0xC7, 0x15, 0x85),
    MIDNIGHTBLUE: rgb(0x19, 0x19, 0x70),
    MINTCREAM: rgb(0xF5, 0xFF, 0xFA),
    MISTYROSE: rgb(0xFF, 0xE4, 0xE1),
    MOCCASIN: rgb(0xFF, 0xE4, 0xB5),
    NAVAJOWHITE: rgb(0xFF, 0xDE, 0xAD),
    NAVY: rgb(0x00, 0x00, 0x80),
    OLDLACE: rgb(0xFD, 0xF5, 0xE6),
    OLIVE: rgb(0x80, 0x80, 0x00),
    OLIVEDRAB: rgb(0x6B, 0x8E, 0x23),
    ORANGE: rgb(0xFF, 0xA5, 0x00),
    ORANGERED: rgb(0xFF, 0x45, 0x00),
    ORCHID: rgb(0xDA, 0x70, 0xD6),
    PALEGOLDENROD: rgb(0xEE, 0xE8, 0xAA),
    PALEGREEN: rgb(0x98, 0xFB, 0x98),
    PALETURQUOISE: rgb(0xAF, 0xEE, 0xEE),
    PALEVIOLETRED: rgb(0xDB, 0x70, 0x93),
    PAPAYAWHIP: rgb(0xFF, 0xEF, 0xD5),
    PEACHPUFF: rgb(0xFF, 0xDA, 0xB9),
    PERU: rgb(0xCD, 0x85, 0x3F),
    PINK: rgb(0xFF, 0xC0, 0xCB),
    PLUM: rgb(0xDD, 0xA0, 0xDD),
    POWDERBLUE: rgb(0xB0, 0xE0, 0xE6),
    PURPLE: rgb(0x80, 0x00, 0x80),
    REBECCAPURPLE: rgb(0x66, 0x33, 0x99),
    RED: rgb(0xFF, 0x00, 0x00),
    ROSYBROWN: rgb(0xBC, 0x8F, 0x8F),
    ROYALBLUE: rgb(0x41, 0x69, 0xE1),
    SADDLEBROWN: rgb(0x8B, 0x45, 0x13),
    SALMON: rgb(0xFA, 0x80, 0x72),
    SANDYBROWN: rgb(0xF4, 0xA4, 0x60),
    SEAGREEN: rgb(0x2E, 0x8B, 0x57),
    SEASHELL: rgb(0xFF, 0xF5, 0xEE),
    SIENNA: rgb(0xA0, 0x52, 0x2D),
    SILVER: rgb(0xC0, 0xC0, 0xC0),
    SKYBLUE: rgb(0x87, 0xCE, 0xEB),
    SLATEBLUE: rgb(0x6A, 0x5A, 0xCD),
    SLATEGRAY: rgb(0x70, 0x80, 0x90),
    SLATEGREY: rgb(0x70, 0x80, 0x90),
    SNOW: rgb(0xFF, 0xFA, 0xFA),
    SPRINGGREEN: rgb(0x00, 0xFF, 0x7F),
    STEELBLUE: rgb(0x46, 0x82, 0xB4),
    TAN: rgb(0xD2, 0xB4, 0x8C),
    TEAL: rgb(0x00, 0x80, 0x80),
    THISTLE: rgb(0xD8, 0xBF, 0xD8),
    TOMATO: rgb(0xFF, 0x63, 0x47),
    TURQUOISE: rgb(0x40, 0xE0, 0xD0),
    VIOLET: rgb(0xEE, 0x82, 0xEE),
    WHEAT: rgb(0xF5, 0xDE, 0xB3),
    WHITE: rgb(0xFF, 0xFF, 0xFF),
    WHITESMOKE: rgb(0xF5, 0xF5, 0xF5),
    YELLOW: rgb(0xFF, 0xFF, 0x00),
    YELLOWGREEN: rgb(0x9A, 0xCD, 0x32),
}


__all__ = [
    "Color",
    "rgba",
    "rgb",
    "hsla",
    "hsl",
    "color",
    "NAMED_COLOR",
    "TRANSPARENT",
] + [name.upper() for name in NAMED_COLOR.keys()]



######################################################################
# Common constants
######################################################################

NORMAL = "normal"
LEFT = "left"
RIGHT = "right"
TOP = "top"
BOTTOM = "bottom"
CENTER = "center"
START = "start"
END = "end"

######################################################################
# Direction
######################################################################

ROW = "row"
COLUMN = "column"

######################################################################
# Visibility
######################################################################

VISIBLE = "visible"
HIDDEN = "hidden"
NONE = "none"

######################################################################
# Text Justification
######################################################################

JUSTIFY = "justify"

######################################################################
# Text Direction
######################################################################

RTL = "rtl"
LTR = "ltr"

######################################################################
# Font family
######################################################################

SYSTEM = "system"
MESSAGE = "message"

SERIF = "serif"
SANS_SERIF = "sans-serif"
CURSIVE = "cursive"
FANTASY = "fantasy"
MONOSPACE = "monospace"

######################################################################
# Font Styling
######################################################################

ITALIC = "italic"
OBLIQUE = "oblique"

FONT_STYLES = {ITALIC, OBLIQUE}

######################################################################
# Font Variant
######################################################################

SMALL_CAPS = "small-caps"

FONT_VARIANTS = {SMALL_CAPS}

######################################################################
# Font boldness
######################################################################

BOLD = "bold"

FONT_WEIGHTS = {BOLD}

######################################################################
# Font Size
######################################################################

SYSTEM_DEFAULT_FONT_SIZE = -1

XX_SMALL = "xx-small"
X_SMALL = "x-small"
SMALL = "small"
MEDIUM = "medium"
LARGE = "large"
X_LARGE = "x-large"
XX_LARGE = "xx-large"
XXX_LARGE = "xxx-large"

ABSOLUTE_FONT_SIZES = {
    XX_SMALL,
    X_SMALL,
    SMALL,
    MEDIUM,
    LARGE,
    X_LARGE,
    XX_LARGE,
    XXX_LARGE,
}

LARGER = "larger"
SMALLER = "smaller"

RELATIVE_FONT_SIZES = {LARGER, SMALLER}

######################################################################
# Colors
######################################################################

TRANSPARENT = "transparent"

ALICEBLUE = "aliceblue"
ANTIQUEWHITE = "antiquewhite"
AQUA = "aqua"
AQUAMARINE = "aquamarine"
AZURE = "azure"
BEIGE = "beige"
BISQUE = "bisque"
BLACK = "black"
BLANCHEDALMOND = "blanchedalmond"
BLUE = "blue"
BLUEVIOLET = "blueviolet"
BROWN = "brown"
BURLYWOOD = "burlywood"
CADETBLUE = "cadetblue"
CHARTREUSE = "chartreuse"
CHOCOLATE = "chocolate"
CORAL = "coral"
CORNFLOWERBLUE = "cornflowerblue"
CORNSILK = "cornsilk"
CRIMSON = "crimson"
CYAN = "cyan"
DARKBLUE = "darkblue"
DARKCYAN = "darkcyan"
DARKGOLDENROD = "darkgoldenrod"
DARKGRAY = "darkgray"
DARKGREY = "darkgrey"
DARKGREEN = "darkgreen"
DARKKHAKI = "darkkhaki"
DARKMAGENTA = "darkmagenta"
DARKOLIVEGREEN = "darkolivegreen"
DARKORANGE = "darkorange"
DARKORCHID = "darkorchid"
DARKRED = "darkred"
DARKSALMON = "darksalmon"
DARKSEAGREEN = "darkseagreen"
DARKSLATEBLUE = "darkslateblue"
DARKSLATEGRAY = "darkslategray"
DARKSLATEGREY = "darkslategrey"
DARKTURQUOISE = "darkturquoise"
DARKVIOLET = "darkviolet"
DEEPPINK = "deeppink"
DEEPSKYBLUE = "deepskyblue"
DIMGRAY = "dimgray"
DIMGREY = "dimgrey"
DODGERBLUE = "dodgerblue"
FIREBRICK = "firebrick"
FLORALWHITE = "floralwhite"
FORESTGREEN = "forestgreen"
FUCHSIA = "fuchsia"
GAINSBORO = "gainsboro"
GHOSTWHITE = "ghostwhite"
GOLD = "gold"
GOLDENROD = "goldenrod"
GRAY = "gray"
GREY = "grey"
GREEN = "green"
GREENYELLOW = "greenyellow"
HONEYDEW = "honeydew"
HOTPINK = "hotpink"
INDIANRED = "indianred"
INDIGO = "indigo"
IVORY = "ivory"
KHAKI = "khaki"
LAVENDER = "lavender"
LAVENDERBLUSH = "lavenderblush"
LAWNGREEN = "lawngreen"
LEMONCHIFFON = "lemonchiffon"
LIGHTBLUE = "lightblue"
LIGHTCORAL = "lightcoral"
LIGHTCYAN = "lightcyan"
LIGHTGOLDENRODYELLOW = "lightgoldenrodyellow"
LIGHTGRAY = "lightgray"
LIGHTGREY = "lightgrey"
LIGHTGREEN = "lightgreen"
LIGHTPINK = "lightpink"
LIGHTSALMON = "lightsalmon"
LIGHTSEAGREEN = "lightseagreen"
LIGHTSKYBLUE = "lightskyblue"
LIGHTSLATEGRAY = "lightslategray"
LIGHTSLATEGREY = "lightslategrey"
LIGHTSTEELBLUE = "lightsteelblue"
LIGHTYELLOW = "lightyellow"
LIME = "lime"
LIMEGREEN = "limegreen"
LINEN = "linen"
MAGENTA = "magenta"
MAROON = "maroon"
MEDIUMAQUAMARINE = "mediumaquamarine"
MEDIUMBLUE = "mediumblue"
MEDIUMORCHID = "mediumorchid"
MEDIUMPURPLE = "mediumpurple"
MEDIUMSEAGREEN = "mediumseagreen"
MEDIUMSLATEBLUE = "mediumslateblue"
MEDIUMSPRINGGREEN = "mediumspringgreen"
MEDIUMTURQUOISE = "mediumturquoise"
MEDIUMVIOLETRED = "mediumvioletred"
MIDNIGHTBLUE = "midnightblue"
MINTCREAM = "mintcream"
MISTYROSE = "mistyrose"
MOCCASIN = "moccasin"
NAVAJOWHITE = "navajowhite"
NAVY = "navy"
OLDLACE = "oldlace"
OLIVE = "olive"
OLIVEDRAB = "olivedrab"
ORANGE = "orange"
ORANGERED = "orangered"
ORCHID = "orchid"
PALEGOLDENROD = "palegoldenrod"
PALEGREEN = "palegreen"
PALETURQUOISE = "paleturquoise"
PALEVIOLETRED = "palevioletred"
PAPAYAWHIP = "papayawhip"
PEACHPUFF = "peachpuff"
PERU = "peru"
PINK = "pink"
PLUM = "plum"
POWDERBLUE = "powderblue"
PURPLE = "purple"
REBECCAPURPLE = "rebeccapurple"
RED = "red"
ROSYBROWN = "rosybrown"
ROYALBLUE = "royalblue"
SADDLEBROWN = "saddlebrown"
SALMON = "salmon"
SANDYBROWN = "sandybrown"
SEAGREEN = "seagreen"
SEASHELL = "seashell"
SIENNA = "sienna"
SILVER = "silver"
SKYBLUE = "skyblue"
SLATEBLUE = "slateblue"
SLATEGRAY = "slategray"
SLATEGREY = "slategrey"
SNOW = "snow"
SPRINGGREEN = "springgreen"
STEELBLUE = "steelblue"
TAN = "tan"
TEAL = "teal"
THISTLE = "thistle"
TOMATO = "tomato"
TURQUOISE = "turquoise"
VIOLET = "violet"
WHEAT = "wheat"
WHITE = "white"
WHITESMOKE = "whitesmoke"
YELLOW = "yellow"
YELLOWGREEN = "yellowgreen"



from ..colors import color


class Choices:
    "A class to define allowable data types for a property"

    def __init__(
        self,
        *constants,
        string=False,
        integer=False,
        number=False,
        color=False,
    ):
        self.constants = set(constants)

        self.string = string
        self.integer = integer
        self.number = number
        self.color = color

        self._options = sorted(str(c).lower().replace("_", "-") for c in self.constants)
        if self.string:
            self._options.append("<string>")
        if self.integer:
            self._options.append("<integer>")
        if self.number:
            self._options.append("<number>")
        if self.color:
            self._options.append("<color>")

    def validate(self, value):
        if self.string:
            try:
                return value.strip()
            except AttributeError:
                pass
        if self.integer:
            try:
                return int(value)
            except (ValueError, TypeError):
                pass
        if self.number:
            try:
                return float(value)
            except (ValueError, TypeError):
                pass
        if self.color:
            try:
                return color(value)
            except ValueError:
                pass
        for const in self.constants:
            if value == const:
                return const

        raise ValueError(f"{value!r} is not a valid value")

    def __str__(self):
        return ", ".join(self._options)



######################################################################
# 2-2025: Backwards compatibility for Toga < 0.5.0
######################################################################

# declaration.py has been split into two modules. Toga < 0.5.0 will look here for
# BaseStyle and Choices.

from warnings import filterwarnings, warn

# Make sure deprecation warnings are shown by default
filterwarnings("default", category=DeprecationWarning)

warn(
    "The travertino.declaration module is deprecated; its contents are now located in "
    "travertino.style and travertino.properties.\n"
    "This error probably means you've updated Travertino to 0.5.0 but are "
    "still using Toga <= 0.4.8; to fix, either update Toga to >= 0.5.0, or "
    "pin Travertino to 0.3.0.",
    DeprecationWarning,
    stacklevel=2,
)

from .properties.choices import Choices  # noqa
from .style import BaseStyle  # noqa



from collections.abc import Sequence

from .choices import Choices
from .immutablelist import ImmutableList


class validated_property:
    def __init__(
        self,
        *constants,
        string=False,
        integer=False,
        number=False,
        color=False,
        initial=None,
    ):
        """Define a simple validated property attribute.

        :param constants: Explicitly allowable values.
        :param string: Are strings allowed as values?
        :param integer: Are integers allowed as values?
        :param number: Are numbers allowed as values?
        :param color: Are colors allowed as values?
        :param initial: The initial value for the property. If the property has not been
            explicitly set, this is what is returned when it's accessed.
        """
        self.choices = Choices(
            *constants, string=string, integer=integer, number=number, color=color
        )
        self.initial = None if initial is None else self.validate(initial)

    def __set_name__(self, owner, name):
        self.name = name
        owner._BASE_PROPERTIES[owner].add(name)
        owner._BASE_ALL_PROPERTIES[owner].add(name)

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        return getattr(obj, f"_{self.name}", self.initial)

    def __set__(self, obj, value):
        if value is self:
            # This happens during autogenerated dataclass __init__ when no value is
            # supplied.
            return

        if value is None:
            raise ValueError(
                "Python `None` cannot be used as a style value; "
                f"to reset a property, use del `style.{self.name}`."
            )

        value = self.validate(value)

        if (current := getattr(obj, f"_{self.name}", None)) is None:
            # If the value has not been explicitly set already, then we always want to
            # assign to the attribute -- even if the value being assigned is identical
            # to the initial value.
            setattr(obj, f"_{self.name}", value)
            if value != self.initial:
                obj.apply(self.name)

        elif value != current:
            setattr(obj, f"_{self.name}", value)
            obj.apply(self.name)

    def __delete__(self, obj):
        try:
            delattr(obj, f"_{self.name}")
        except AttributeError:
            pass
        else:
            obj.apply(self.name)

    @property
    def _name_if_set(self):
        return f" {self.name}" if hasattr(self, "name") else ""

    def validate(self, value):
        try:
            return self.choices.validate(value)
        except ValueError:
            raise ValueError(
                f"Invalid value {value!r} for property{self._name_if_set}; "
                f"Valid values are: {self.choices}"
            )

    def is_set_on(self, obj):
        return hasattr(obj, f"_{self.name}")


class list_property(validated_property):
    def validate(self, value):
        if isinstance(value, str):
            value = [value]
        elif not isinstance(value, Sequence):
            raise TypeError(
                f"Value for list property{self._name_if_set} must be a sequence."
            )

        if not value:
            name = getattr(self, "name", "prop_name")
            raise ValueError(
                "List properties cannot be set to an empty sequence; "
                f"to reset a property, use del `style.{name}`."
            )

        # This could be a comprehension, but then the error couldn't specify which value
        # is at fault.
        result = []
        for item in value:
            try:
                item = self.choices.validate(item)
            except ValueError:
                raise ValueError(
                    f"Invalid item value {item!r} for list "
                    f"property{self._name_if_set}; Valid values are: {self.choices}"
                )
            result.append(item)

        return ImmutableList(result)



from ..constants import BOTTOM, LEFT, RIGHT, TOP


class directional_property:
    DIRECTIONS = [TOP, RIGHT, BOTTOM, LEFT]
    ASSIGNMENT_SCHEMES = {
        #   T  R  B  L
        1: [0, 0, 0, 0],
        2: [0, 1, 0, 1],
        3: [0, 1, 2, 1],
        4: [0, 1, 2, 3],
    }

    def __init__(self, name_format):
        """Define a property that proxies for top/right/bottom/left alternatives.

        :param name_format: The format from which to generate subproperties. "{}" will
            be replaced with "_top", etc.
        """
        self.property_names = [
            name_format.format(f"_{direction}") for direction in self.DIRECTIONS
        ]

    def __set_name__(self, owner, name):
        self.name = name
        owner._BASE_ALL_PROPERTIES[owner].add(self.name)

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        return tuple(obj[name] for name in self.property_names)

    def __set__(self, obj, value):
        if value is self:
            # This happens during autogenerated dataclass __init__ when no value is
            # supplied.
            return

        if not isinstance(value, tuple):
            value = (value,)

        if order := self.ASSIGNMENT_SCHEMES.get(len(value)):
            for name, index in zip(self.property_names, order):
                obj[name] = value[index]
        else:
            raise ValueError(
                f"Invalid value for '{self.name}'; value must be a number, or a 1-4 "
                f"tuple."
            )

    def __delete__(self, obj):
        for name in self.property_names:
            del obj[name]

    def is_set_on(self, obj):
        return any(name in obj for name in self.property_names)
